/**
 * Crée une instance de Webos.Cmd, representant une commande.
 * @param {Object} options Les options de la commande.
 * @since 1.0 alpha 1
 * @constructor
 */
Webos.Cmd = function WCmd(options) {
	this.cmdText = options.cmd; //Commande complete
	this.cmd = this.cmdText.split(' ').shift(); //Nom de la commande
	this.terminal = options.terminal; //Terminal d'ou la commande sera lancee
	
	//On appelle la classe parente
	Webos.Process.call(this, {
		args: Webos.Arguments.parse(options.cmd)
	});
};
Webos.Cmd.prototype = {
	/**
	 * Recuperer le terminal d'ou la commande est lancee.
	 * @returns {Webos.Terminal} Le terminal.
	 */
	getTerminal: function() {
		return this.terminal;
	},
	/**
	 * Lancer la commande.
	 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que la commande aura ete executee.
	 */
	run: function(callback) {
		callback = Webos.Callback.toCallback(callback);
		var that = this;
		
		new Webos.ServerCall({
			'class': 'CmdController',
			'method': 'execute',
			'arguments': { 'cmd': this.cmdText, 'terminal': this.terminal.getId() }
		}).load([function(response) {
			var out = response.getAllChannels();
			if (out) {
				that.getTerminal().echo(out);
			}

			var data = response.getData();
			if (data.script) { //Si il y a du code JS a executer, on l'execute
				var auth = new Webos.Authorizations();
				for (var index in data.authorizations) {
					auth.add(data.authorizations[index]);
				}

				Webos.Process.call(that, {
					pid: data.pid,
					key: data.key,
					authorizations: auth,
					fn: data.script
				});
				Webos.Cmd.uber.run.call(that);

				callback.success(that);
			} else {
				this.stop();

				callback.success(that);
			}
		}, function(response) {
			var out = response.getAllChannels();
			if (out) {
				that.getTerminal().echo(out);
			}

			callback.error(response);
		}]);
	}
};
Webos.inherit(Webos.Cmd, Webos.Process); //Heritage de Webos.Process

/**
 * Executer une commande.
 * @param {String} cmd La commande a executer.
 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que la commande aura ete executee.
 * @static
 */
Webos.Cmd.execute = function(cmd, callback) {
	callback = Webos.Callback.toCallback(callback);
	
	var terminal = Webos.Terminal.create();
	terminal.enterCmd(cmd, callback);
};

/**
 * Crée une instance de Webos.Terminal, representant un terminal.
 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que le terminal aura ete initialise.
 * @since 1.0 alpha 1
 * @constructor
 */
Webos.Terminal = function WTerminal() {
	this._data = {
		location: '~'
	};
	this._initialized = false;
	this._output = '';
	
	this.id = Webos.Terminal.register(this);

	Webos.Observable.call(this);
};
Webos.Terminal.prototype = {
	/**
	 * Recuperer l'identifiant du terminal.
	 * @returns {Number} L'identifiant du terminal.
	 */
	getId: function() {
		return this.id;
	},
	/**
	 * Recuperer toutes les donnees du terminal.
	 * @returns Toutes les donnees.
	 */
	data: function() {
		return this._data;
	},
	/**
	 * Recuperer une donnee du terminal.
	 * @param {String} key La clef de la donnee.
	 * @returns La valeur correspondant a la clef.
	 */
	get: function(key) {
		return this._data[key];
	},
	/**
	 * Set this terminal's location.
	 * @param {String} path The new location.
	 */
	setLocation: function (path) {
		this._data['location'] = path;
	},
	/**
	 * Recuperer un chemin absolu depuis un chemin relatif par rapport au dossier courant.
	 * @param {String} path Le chemin relatif par rapport au dossier courant.
	 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que le chemin sera converti.
	 * @deprecated Use absolutePath().
	 */
	relativePath: function(path, callback) {
		callback.success(this.get('location')+'/'+path);
	},
	/**
	 * Recuperer un chemin absolu depuis un chemin relatif par rapport au dossier courant.
	 * @param {String} path Le chemin relatif par rapport au dossier courant.
	 */
	absolutePath: function(path) {
		var absPath = path;
		if (path.substr(0, 1) != '/' && path.substr(0, 2) != '~/' && path != '~') { //Relative path
			absPath = this.get('location')+'/'+path;
		}

		return absPath;
	},
	/**
	 * Output some content.
	 * @param  {string} contents The content.
	 */
	echo: function(contents) {
		contents = String(contents).replace(/\n/g, '<br />').replace(/(^ | $)/g, '&nbsp;');
		this._output += contents;
		this.notify('echo', {
			contents: contents,
			output: this._output,
			cmd: this.cmd
		});
	},
	/**
	 * Prompt some question.
	 * @param  {Function} callback The callback.
	 * @param  {object}   options  Options, which are mainly "label" and "type".
	 */
	prompt: function(callback, options) {
		var that = this;

		if (this._promptCallback) {
			return;
		}

		var originalOptions = options;
		options = $.extend({
			label: '',
			type: 'text'
		}, options);

		if (options.type == 'yn') {
			options.label += ' [Y/n] ';
		}

		var triggerResult = function (val) {
			callback(val);
		};

		this._promptCallback = function(e, el) {
			if (e.type == 'keydown') {
				if (e.keyCode == 13) {
					e.preventDefault();
					that._promptCallback = null;

					var val = $(el).val();
					$(el).replaceWith(val);

					that.echo('\n');

					switch (options.type) {
						case 'yn':
							if (val == 'Y' || val == 'n') {
								triggerResult(val);
							} else {
								if (originalOptions.label.indexOf('\n') != 0) {
									originalOptions.label = '\n' + originalOptions.label;
								}
								that.prompt(callback, originalOptions);
							}
							break;
						case 'text':
						default:
							triggerResult(val);
					}
				}
			}
		};

		var callbackName = 'Webos.Terminal._list['+this.getId()+']._promptCallback'
		options.label += '<input type="text" onkeydown="'+callbackName+'(event || window.event, this);"';

		if (options.type == 'yn') {
			options.label += ' size="2" maxlength="1"';
		}

		options.label += '/>';

		this.echo(options.label);
	},
	/**
	 * Executer une commande dans le terminal.
	 * @param {String} cmd La commande a executer.
	 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que la commande aura ete executee.
	 * @returns {Webos.Cmd} La commande.
	 */
	enterCmd: function(cmd, callback) {
		this._output = '';
		this.cmd = new Webos.Cmd({
			cmd: cmd,
			terminal: this
		});
		this.cmd.run(Webos.Callback.toCallback(callback));
		return this.cmd;
	},
	_refreshUserData: function(username, callback) {
		callback = Webos.Callback.toCallback(callback);

		if (!username) {
			callback.success({
				root: false
			});
			return;
		}

		Webos.User.getByUsername(username, [function(user) {
			if (!user) {
				var data = {
					root: false
				};
				callback.success(data);
				return;
			}
			
			user.authorizations([function(auth) {
				var model = auth.model();
				var data = {
					root: (model == 'admin')
				};
				callback.success(data);
			}, callback.error]);
		}, callback.error]);
	},
	/**
	 * Rafraichir les donnees sur le terminal.
	 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que les donnees auront ete rafraichies.
	 */
	refreshData: function(callback) {
		callback = Webos.Callback.toCallback(callback);
		
		var that = this;
		
		if (!this._initialized) {
			this.init(callback);
		} else {
			new Webos.ServerCall({
				'class': 'TerminalController',
				method: 'getPromptData',
				arguments: { 'terminal': this.getId() }
			}).load([function(response) {
				var data = response.getData();
				if (!data.host) {
					data.host = window.location.host;
				}
				that._data = data;

				that._refreshUserData(that._data.username, [function(data) {
					that._data = $.extend({}, that._data, data);
					callback.success(that);
				}, function() {
					callback.success(that);
				}]);
			}, callback.error]);
		}
	},
	/**
	 * Initialiser le terminal.
	 * @param {Webos.Callback} callback La fonction de rappel qui sera appelee une fois que le terminal aura ete initialise.
	 */
	init: function(callback) {
		callback = Webos.Callback.toCallback(callback);
		
		var that = this;
		
		new Webos.ServerCall({
			'class': 'TerminalController',
			method: 'register',
			arguments: { terminal: this.getId() }
		}).load([function(response) {
			var data = response.getData();
			if (!data.host) {
				data.host = window.location.host;
			}
			that._data = data;

			that._initialized = true;
			
			that._refreshUserData(that._data.username, [function(data) {
				that._data = $.extend({}, that._data, data);
				callback.success(that);
			}, function() {
				callback.success(that);
			}]);
		}, callback.error]);
	}
};
Webos.inherit(Webos.Terminal, Webos.Observable);

/**
 * La liste des terminaux.
 * @private
 * @static
 */
Webos.Terminal._list = [];

/**
 * Ajouter un terminal a la liste interne des terminaux.
 * @param {Webos.Terminal} terminal Le terminal a ajouter.
 * @returns {Number} L'identifiant du terminal dans la liste.
 * @static
 */
Webos.Terminal.register = function(terminal) {
	return Webos.Terminal._list.push(terminal) - 1;
};

/**
 * Recuperer un terminal a partir de son identifiant.
 * @param {Number} id L'identifiant du terminal.
 * @returns {Webos.Terminal} Le terminal.
 * @static
 */
Webos.Terminal.get = function(id) {
	return Webos.Terminal._list[id];
};

Webos.Terminal.create = function() {
	return new Webos.Terminal();
};